from .dkn import DKN as NN__DKN
from .lstur import LSTUR as NN__LSTUR
from .naml import NAML as NN__NAML
from .npa import NPA as NN__NPA
from .nrms import NRMS as NN__NRMS
from .cnn import CNN as NN__CNN

from .autoencoder import Autoencoder as NN__AUTOENCODER

from .vae import VAE as NN__VAE
from .transformer import TransformerModel as NN__TransformerModel
from .rnn import RNNModel as NN__RNNModel
from .AITM import AITM as TRA_AITM
from .DSSM import DSSM as TRA_DSSM
from .MIND import MIND as TRA_MIND   #transformer class is prefixed as "TRA" abbr
from .TDM import TDM as TRA_TDM
from .WidenDeep import WidenDeep as TRA_WIDENDEEP
from .Word2Vec import Word2Vec as TRA_WORD2VEC
from .Youtube_dnn import YoutubeDNN as TRA_YOUTUBEDNN   
