from .sasrec_base import SASRecBase as NN_SASREC
from .rnn_sequential_recommendation_base import RNN_sequential_recommendation_base as NN_RNN_SEQUENTIAL_RECOMMENDATION
from .self_supervised_learning_cf_base import SelfSupervisedLearningCFBase as NN_SELF_SUPERVISED_LEARNING_CF
from .xdeepfm_base import XDeepFMBase as NN_XDEEPFM
from .ncf_base import NCF_base as NN_NCF
from .autoencoder_cf_base import AutoencoderCFBase as NN_AUTOENCODER_CF
from .hybrid_deep_learning_base import HybridDeepLearningBase as NN_HYBRID_DEEP_LEARNING
from .nextitnet_base import NextItNetBase as NN_NEXTITNET
from .ssept_base import SSEPTBase as NN_SSEPT
from .Bert4Rec_base import Bert4Rec_base as NN_BERT4REC
from .gnn_ufilter_base import GNN_ufilter_base as NN_GNN
from .gan_ufilter_base import GAN_ufilter_base as NN_GAN
from .transformer_ufilter_base import TransformerUFilterBase as NN_TRANSFORMER
from .variational_autoencoder_ufilter_base import VariationalAutoencoderUFilterBase as NN_VARIATIONAL_AUTOENCODER
from .slirec_base import SLiRecBase as NN_SLIREC
from .SSR_base import SSR_base as NN_SSR
from .deep_mf_base import DeepMF_base as NN_DEEP_MF
from .bivae_base import BVAE_base as NN_BVAE
from .neural_mf_base import NeuralMFBase as NN_NEURAL_MF
from .FM_base import FM_base as NN_FM
from .DeepFM_base import DeepFM_base as NN_DEEPFM
from .PNN_base import PNN_base as NN_PNN
from .AFM_base import AFM_base as NN_AFM
from .BST_base import BST_base as NN_BST
from .NFM_base import NFM_base as NN_NFM
from .ENSFM_base import ENSFM_base as NN_ENSFM
from .DeepRec_base import DeepRec_base as NN_DEEPREC
from .DeepFEFM_base import DeepFEFM_base as NN_DEEPFEFM
from .GNN_base import GNN_base as NN_GNN_BASE
from .PLE_base import PLE_base as NN_PLE
from .gru_ufilter_base import GRU_ufilter_base as NN_GRU
from .FFM_base import FFM_base as NN_FFM
from .FGCNN_base import FGCNN_base as NN_FGCNN
from .neural_ufilter_base import NeuralCollaborativeFilteringBase as NN_NEURAL_CF
from .RALM_base import RALM_base as NN_RALM
from .AutoFI_base import AutoFI_base as NN_AUTOFI
from .ESMM_base import ESMM_base as NN_ESMM
from .ESCMM_base import ESCMM_base as NN_ESCMM
from .DMR_base import DMR_base as NN_DMR
from .caser_base import Caser_base as NN_CASER
from .ListWise_base import ListWise_base as NN_LISTWISE
from .DIFM_base import DIFM_base as NN_DIFM
from .DeepCrossing_base import DeepCrossing_base as NN_DEEPCROSSING
from .DCN_base import DCN_base as NN_DCN
from .MMOE_base import MMOE_base as NN_MMOE
from .Fibinet_base import Fibinet_base as NN_FIBINET
from .DLRM_base import DLRM_base as NN_DLRM
from .AutoInt_base import AutoInt_base as NN_AUTOINTER
from .MAML_base import MAML_base as NN_MAML
from .GateNet_base import GateNet_base as NN_GATENET
from .DIEN_base import DIEN_base as NN_DIEN
from .FLEN_base import FLEN_base as NN_FLEN
from .TiSAS_base import TiSASBase as NN_TISAS

__all__ = [
    "NN_SASREC",
    "NN_RNN_SEQUENTIAL_RECOMMENDATION",
    "NN_SELF_SUPERVISED_LEARNING_CF",
    "NN_XDEEPFM",
    "NN_NCF",
    "NN_AUTOENCODER_CF",
    "NN_HYBRID_DEEP_LEARNING",
    "NN_NEXTITNET",
    "NN_SSEPT",
    "NN_BERT4REC",
    "NN_GNN",
    "NN_GAN",
    "NN_TRANSFORMER",
    "NN_VARIATIONAL_AUTOENCODER",
    "NN_SLIREC",
    "NN_SSR",
    "NN_DEEP_MF",
    "NN_BVAE",
    "NN_NEURAL_MF",
    "NN_FM",
    "NN_DEEPFM",
    "NN_PNN",
    "NN_AFM",
    "NN_BST",
    "NN_NFM",
    "NN_ENSFM",
    "NN_DEEPREC",
    "NN_DEEPFEFM",
    "NN_GNN_BASE",
    "NN_PLE",
    "NN_GRU",
    "NN_FFM",
    "NN_FGCNN",
    "NN_NEURAL_CF",
    "NN_RALM",
    "NN_AUTOFI",
    "NN_ESMM",
    "NN_ESCMM",
    "NN_DMR",
    "NN_CASER",
    "NN_LISTWISE",
    "NN_DIFM",
    "NN_DEEPCROSSING",
    "NN_DCN",
    "NN_MMOE",
    "NN_FIBINET",
    "NN_DLRM",
    "NN_AUTOINTER",
    "NN_MAML",
    "NN_GATENET",
    "NN_DIEN",
    "NN_FLEN",
    "NN_DIN",
    "NN_TISAS",
]

# init is in maintenance mode
            